local print = function(msg) if msg then DEFAULT_CHAT_FRAME:AddMessage(msg) end end
--local pdebug = function(msg) if msg then table.insert(UABDebug, msg) end end
--local pdebug = function(msg) print(msg) end
local pdebug = function(msg) end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabsquare
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabsquare = UnitActionBars.aceOO.Class(UnitActionBars.layoutManager.ILayout, UnitActionBars.aceEvent)

function UnitActionBars.layouts.uabsquare.GetLayoutName()
    return "UABSquare"
end

function UnitActionBars.layouts.uabsquare.GetDisplayName()
    return "Square"
end

function UnitActionBars.layouts.uabsquare.GetConfigTemplate()
    return "UAB_SquareLayoutOptionsTemplate"
end
     
function UnitActionBars.layouts.uabsquare.prototype:init(configFrame, sampleFrame, settings)
    UnitActionBars.layouts.uabsquare.super.prototype.init(self)

    self.name = "uabsquare: "..configFrame:GetName().." : "..sampleFrame:GetName()
    
    pdebug(self.name..":init")
    
    self:RegisterEvent("UnitActionBars_CreateGroupUnit", "CreateGroupUnit")

    self.configFrame = configFrame
    self.sampleFrame = sampleFrame
    configFrame.layout = self
    
    self.controllers = {}
    self.settings = UnitActionBars.layouts.uabsquare.settings:new(configFrame, settings)
    self.sampleController = UnitActionBars.layouts.uabsquare.controller:new(self.settings, "sample", 1, sampleFrame, 1)

    configFrame:SetScript("OnShow", function(self) self.layout:ShowSample() end)
    configFrame:SetScript("OnHide", function(self) self.layout:HideSample() end)
end

function UnitActionBars.layouts.uabsquare.prototype:ShowSample()
    pdebug(self.name..":ShowSample")
    
    self.sampleController:EnableActionSet()
end

function UnitActionBars.layouts.uabsquare.prototype:HideSample()
    pdebug(self.name..":HideSample")

    self.sampleController:DisableActionSet()    
end

function UnitActionBars.layouts.uabsquare.prototype:CreateGroupUnit(name, id, header)
    pdebug(self.name..":CreateGroupUnit")
    
    self.controllers[name..id] = UnitActionBars.layouts.uabsquare.controller:new(self.settings, name, id, header)
end

function UnitActionBars.layouts.uabsquare.prototype:GetSettings()
    pdebug(self.name..":GetSettings")
    
    self.settings:GetValues()
end

function UnitActionBars.layouts.uabsquare.prototype:DefaultSettings(bypassNotification)
    pdebug(self.name..":DefaultSettings")
    
    self.settings:Defaults(bypassNotification)
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabsquare.controller
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabsquare.controller = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

UnitActionBars.layouts.uabsquare.controller.rowButtonStart = { 1, 9 }
UnitActionBars.layouts.uabsquare.controller.rowButtons = { 8, 16 }
UnitActionBars.layouts.uabsquare.controller.buttonOrder = {
    row1 = { "l2", "u2l2", "u2", "u2r2", "r2", "d2r2", "d2", "d2l2" },
    row2 = { "l4", "u2l4", "u4l4", "u4l2", "u4", "u4r2", "u4r4", "u2r4", "r4", "d2r4", "d4r4", "d4r2", "d4", "d4l2", "d4l4", "d2l4" },
}

UnitActionBars.layouts.uabsquare.controller.buttonOrderSequence = {
    [uab.constants.placement.left] = { 1, 1 },
    [uab.constants.placement.topleft] = { 2, 2 },
    [uab.constants.placement.top] = { 3, 5 },
    [uab.constants.placement.topright] = { 4, 8 },
    [uab.constants.placement.right] = { 5, 9 },
    [uab.constants.placement.bottomright] = { 6, 10 },
    [uab.constants.placement.bottom] = { 7, 13 },
    [uab.constants.placement.bottomleft] = { 8, 16 }
}

function UnitActionBars.layouts.uabsquare.controller.prototype:init(settings, name, id, header, sample)
    UnitActionBars.layouts.uabsquare.controller.super.prototype.init(self)

    self.name = "uabsquarecontroller: "..name.." : "..id
    if sample then
        self.name = self.name .. "sample"
    end
    
    pdebug(self.name..":init")
    
    self.id = id
    self.settings = settings
    self.sample = sample
    self.header = header
    self.groupname = name
    self.reposition = 1
    self.revisual = 1
    self.enabled = nil

    local actionsetname
    
    if sample then
        actionsetname = header:GetName().."UABSquareSample"

        self.header = CreateFrame("Frame", actionsetname.."Frame", header)
        self.header:SetHeight(header:GetWidth())
        self.header:SetWidth(header:GetHeight())
        self.header:SetPoint("CENTER")
    end
    
    self.actionset = UnitActionBars.actionsets.standard:GetActionSet(self.header, id, actionsetname)
    self.actionset:SetController(self)

    self:RegisterEvent("UnitActionBars_UABSquare_SettingsChanged", "SettingsChanged")
    self:RegisterEvent("UnitActionBars_"..self.groupname.."_ChangeVisualActionEffect", "ChangeVisualEffect")
    self:RegisterEvent("UnitActionBars_UABSquare_Enable", "EnableActionSet")
    self:RegisterEvent("UnitActionBars_UABSquare_Disable", "DisableActionSet")
    
    if not sample then
        self:RegisterEvent("UnitActionBars_"..name.."_SetAction", "SetAction")
        self:RegisterEvent("UnitActionBars_"..name.."_RemoveAction", "RemoveAction")
    end

    self.actionset:SetInUse()
    if self.sample then
        self.header:Hide()
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:SettingsChanged(settings)
    pdebug(self.name..":SettingsChanged")
    
    self.settings = settings
    self.reposition = 1
    self:UpdatePositions()
end

function UnitActionBars.layouts.uabsquare.controller.prototype:ChangeVisualEffect(id, effects)
    pdebug(self.name..":ChangeVisualEffect")
    
    self:TriggerEvent("UnitActionBars_"..self.actionset:ActionSetName().."_ChangeVisualActionEffect"..id, effects)
end

function UnitActionBars.layouts.uabsquare.controller.prototype:EnableActionSet()
    pdebug(self.name..":EnableActionSet")

    if not self.enabled then
        self.enabled = 1
        self.actionset:SetController(self)
        self.actionset:SetInUse(1)
        self.reposition = 1
        self.revisual = 1
        self:UpdateAll()
        
        if self.sample then
            self.header:Show()
        end
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:DisableActionSet()
    pdebug(self.name..":DisableActionSet")

    if self.enabled then
        self.enabled = nil    
        self.actionset:SetInUse()
        if self.sample then
            self.header:Hide()
        end
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:SetAction(number, action)
    pdebug(self.name..":SetAction")
    
    self.reposition = 1
    self.revisual = 1
    self.actionset:SetAction(number, action)
end

function UnitActionBars.layouts.uabsquare.controller.prototype:RemoveAction(number)
    pdebug(self.name..":RemoveAction")
    
    if self.actionset:ActionButton(number) then
        self.reposition = 1
        self.revisual = 1
        self.actionset:RemoveAction(number)
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:UpdateAll()
    pdebug(self.name..":UpdateAll")
    
    self:UpdatePositions()
    self:UpdateVisualEffects()
end

function UnitActionBars.layouts.uabsquare.controller.prototype:UpdatePositions()
    if self.reposition and self.actionset:IsVisible() then
        self.reposition = nil
        
        -- repositioning logic goes here
        self.actionset:ClearPositions()        

        self:PositionRow(1)
        self:PositionRow(2)
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:PositionRow(row)
    local order = UnitActionBars.layouts.uabsquare.controller.buttonOrder["row"..row]
    local startpos = self.settings.startLocation
    local buttonIdx = UnitActionBars.layouts.uabsquare.controller.rowButtonStart[row]
    local sequence = UnitActionBars.layouts.uabsquare.controller.buttonOrderSequence

    if self.settings.growClockwise == "true" then
        for idx = sequence[startpos][row], UnitActionBars.layouts.uabsquare.controller.rowButtons[row], 1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
        
        local endpos = sequence[startpos][row] - 1
        
        for idx = 1, endpos, 1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
    else
        for idx = sequence[startpos][row], 1, -1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
        
        local endpos = sequence[startpos][row] + 1
        
        for idx = UnitActionBars.layouts.uabsquare.controller.rowButtons[row], endpos, -1 do
            self:PositionButton(order[idx], buttonIdx, self.actionset:ActionButton(buttonIdx), self.actionset:CancelButton())
            buttonIdx = buttonIdx + 1
        end
    end
end

function UnitActionBars.layouts.uabsquare.controller.prototype:PositionButton(orderCode, btnidx, btn, center)
    if not btn then return end
    
    local vMod = 0
    local hMod = 0

    tmpCode = orderCode        
    if string.sub(orderCode, 1, 1) == "u" then
        vMod = tonumber(string.sub(orderCode, 2, 2)) * 0.5
        tmpCode = string.sub(orderCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "d" then
        vMod = tonumber(string.sub(tmpCode, 2, 2)) * -0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "r" then
        hMod = tonumber(string.sub(tmpCode, 2, 2)) * 0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    if string.sub(tmpCode, 1, 1) == "l" then
        hMod = tonumber(string.sub(tmpCode, 2, 2)) * -0.5
        tmpCode = string.sub(tmpCode, 3)
    end

    local x = (UnitActionBars.actionbuttons.standard.buttonWidth + 1) * hMod
    local y = (UnitActionBars.actionbuttons.standard.buttonHeight + 1) * vMod
    
    btn:SetPoint("CENTER", center, "CENTER", x, y)
end

function UnitActionBars.layouts.uabsquare.controller.prototype:UpdateVisualEffects()
    if self.revisual and self.actionset:IsVisible() then
        self.revisual = nil
    
        -- apply all the visual properties based on user selections (low alpha, highlight border, blink)
        -- for each button control in the actionset

        --self.actionset:UpdateVisualEffects()        
    end
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabsquare.settings
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabsquare.settings = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

function UnitActionBars.layouts.uabsquare.settings.StartLocationDisplayValue(value)
    if value == uab.constants.placement.top then
        return "Top"
    elseif value == uab.constants.placement.topright then
        return "Top Right"
    elseif value == uab.constants.placement.topleft then
        return "Top Left"
    elseif value == uab.constants.placement.bottom then
        return "Bottom"
    elseif value == uab.constants.placement.bottomright then
        return "Bottom Right"
    elseif value == uab.constants.placement.bottomleft then
        return "Bottom Left"
    elseif value == uab.constants.placement.left then
        return "Left"
    else
        return "Right"
    end
end

function UnitActionBars.layouts.uabsquare.settings.prototype:init(configFrame, settings)
    UnitActionBars.layouts.uabsquare.settings.super.prototype.init(self)

    self.name = "uabsquaresettings: "..configFrame:GetName()

    pdebug(self.name..":init")
    
    self.frame = configFrame
	
    self.growClockwise = (settings and settings.growClockwise) or "true"
    self.startLocation = (settings and settings.startLocation) or uab.constants.placement.top

    self.growDirectionFrame = getglobal(self.frame:GetName().."GrowDirection")
    self.growDirectionFrame.settings = self
	
    getglobal(self.growDirectionFrame:GetName().."Text"):SetText("Add buttons clockwise")
	self.growDirectionFrame:SetScript("OnClick", function(self, button, down) self.settings:GrowDirectionButtonClicked(self) end)

    self.startLocationFrame = getglobal(self.frame:GetName().."StartLocation")
    self.startLocationFrame.settings = self
	
    UnitActionBars.layouts.uabsquare.settings.InsideInit = 1
    UnitActionBars.layouts.uabsquare.settings.InitDropdown = self
    
    UIDropDownMenu_Initialize(self.startLocationFrame,
        function()
            if UnitActionBars.layouts.uabsquare.settings.InsideInit then
                UnitActionBars.layouts.uabsquare.settings.InitDropdown:StartLocationListInitialize()
            else
                this:GetParent().settings:StartLocationListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.startLocationFrame)
	UIDropDownMenu_SetButtonWidth(24, self.startLocationFrame)
	UIDropDownMenu_JustifyText("LEFT", self.startLocationFrame)

    UnitActionBars.layouts.uabsquare.settings.InsideInit = nil
    UnitActionBars.layouts.uabsquare.settings.InitDropdown = nil
    
	self:SetGrowClockwise(self.growClockwise)
	self:SetStartLocation(self.startLocation)
    
    self:RegisterEvent("UnitActionBars_ProfileChanged", "ProfileChanged")    
    self:TriggerEvent("UnitActionBars_UABSquare_SettingsChanged", self)
end

function UnitActionBars.layouts.uabsquare.settings.prototype:ProfileChanged()
    local settings = UnitActionBars.db.profile.layoutSettings.UABOval
    
    self.growClockwise = (settings and settings.growClockwise) or "true"
    self.startLocation = (settings and settings.startLocation) or uab.constants.placement.top

	self:SetGrowClockwise(self.growClockwise)
	self:SetStartLocation(self.startLocation)
end

function UnitActionBars.layouts.uabsquare.settings.prototype:Defaults(bypassNotification)
    pdebug(self.name..":Defaults")
    
    self.growClockwise = "true"
    self.startLocation = uab.constants.placement.top

    UnitActionBars.db.profile.layoutSettings.UABSquare = self:GetValues()
    
    if not bypassNotification then
        self:TriggerEvent("UnitActionBars_UABSquare_SettingsChanged", self)
    end
end

function UnitActionBars.layouts.uabsquare.settings.prototype:StartLocationListInitialize()
    pdebug(self.name..":StartLocationListInitialize")
    
    local index = 1
    local selected = nil

    local options = {}

    table.insert(options, {	text = "Top Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.topleft }})
    table.insert(options, {	text = "Top",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.top }})
    table.insert(options, {	text = "Top Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.topright }})
    table.insert(options, {	text = "Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.right }})
    table.insert(options, {	text = "Bottom Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottomright }})
    table.insert(options, {	text = "Bottom",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottom }})
    table.insert(options, {	text = "Bottom Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.bottomleft }})
    table.insert(options, {	text = "Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectStartLocation", arg1 = uab.constants.placement.left }})

    for index, data in pairs(options) do
        UIDropDownMenu_AddButton(data)
        if self.startLocation == data.arg2.arg1 then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.startLocationFrame, selected)
end

function UnitActionBars.layouts.uabsquare.settings.prototype:SelectStartLocation(selector, selection)
    pdebug(self.name..":SelectStartLocation")
    
    UIDropDownMenu_SetSelectedID(self.startLocationFrame, selector:GetID())
    self:SetStartLocation(selection)
end

function UnitActionBars.layouts.uabsquare.settings.prototype:SetStartLocation(value)
    pdebug(self.name..":SetStartLocation")
    
    local notify = (value ~= self.startLocation)
        
    self.startLocation = value
    UIDropDownMenu_SetText(UnitActionBars.layouts.uabsquare.settings.StartLocationDisplayValue(value), self.startLocationFrame)
    
    UnitActionBars.db.profile.layoutSettings.UABSquare = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABSquare_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabsquare.settings.prototype:GrowDirectionButtonClicked(frame)
    pdebug(self.name..":GrowDirectionButtonClicked")
    
    self:SetGrowClockwise(frame:GetChecked())
end

function UnitActionBars.layouts.uabsquare.settings.prototype:SetGrowClockwise(value)
    pdebug(self.name..":SetGrowClockwise")
    
    local newval
    
    if value and (tonumber(value) == 1 or value == "true") then
        newval = "true"
    else
        newval = "false"
    end
    
    local notify = (newval ~= self.growClockwise)
        
    self.growClockwise = newval
    self.growDirectionFrame:SetChecked(self.growClockwise == "true")

    UnitActionBars.db.profile.layoutSettings.UABSquare = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABSquare_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabsquare.settings.prototype:GetValues()
    pdebug(self.name..":GetValues")
    
    local values = {}
    
    values.growClockwise = self.growClockwise
    values.startLocation = self.startLocation
    
    return values
end

UnitActionBars.layoutManager:Register( UnitActionBars.layouts.uabsquare )
